#include <includes.h>


#define NAME "Web sucker"
#define DESC "\
This plugin mirrors the remote web site\n\
and warns the user if it has found : \n\
a) file listings for directory which have\n\
   no index\n\
b) if the server could not find some referenced\n\
   HTML links.\n\
In addition to that, the list of links is copied\n\
in memory and may be used by some other plugins.\n\
Risk factor: low/medium"

#define SUMM "mirrors the remote website"
#define COPYRIGHT "distributed under the GPL" 

struct file_infos {
  long size;
  int error;
  int html;
  char * hostname;
  char * name;
  char * data;
  } ;
  
  
int http_get_file_info(struct arglist * , char *, char *, int *, long *, int *);
int http_get_path_from_file(char *, char **);
int http_link_get_hostname(char *, char **);
int http_get_file(struct arglist *, struct file_infos *);
int http_extract_links(struct arglist * , char *, char *, struct file_infos *);
int http_filter_link(char *);
void http_remove_double_dot(char **, char *);
int http_data_is_directory_listing(char *);

PlugExport int plugin_init(struct arglist *desc);
PlugExport int plugin_init(struct arglist *desc)
{
   plug_set_name(desc, NAME);
   plug_set_description(desc, DESC);
   plug_set_summary(desc, SUMM);
   plug_set_copyright(desc, COPYRIGHT);
   plug_set_category(desc, ACT_GATHER_INFO);
   plug_set_family(desc, "Misc.");
   return(0);
}


PlugExport int plugin_run(struct arglist * env);
PlugExport int plugin_run(struct arglist * env)
{
 struct file_infos finfos;
 char* path;
 struct arglist * links = emalloc(sizeof(struct arglist));
 struct arglist * directory_listing = emalloc(sizeof(struct arglist));
 struct arglist * error_404 = emalloc(sizeof(struct arglist));
 int directory_listings = 0;
 long report_length_404 = 0;
 long report_length_listings = 0;
 struct arglist * export = emalloc(sizeof(struct arglist));
 int errors_404 = 0;
 struct arglist * l = links;
 struct arglist * first = emalloc(sizeof(struct arglist));
 arg_add_value(first, "parent", ARG_STRING, 1, "/");
 arg_add_value(first, "error_code", ARG_INT, sizeof(int), (void *)0);
 arg_add_value(links, "/", ARG_ARGLIST, -1, first);
 
 finfos.hostname = (char *)plug_get_hostname(env);
 while(l && l->next)
 {
 struct arglist * infos = l->value;
 int code = 0;
 
 finfos.data = NULL;
 if(infos)code = (int)arg_get_value(infos, "error_code");
 if(!code)
 {
 finfos.name = l->name;
 http_get_file(env, &finfos);
 if((finfos.error == 200)&&(finfos.data))
 {
 http_get_path_from_file(finfos.name, &path);
 http_extract_links(links, path, finfos.data, &finfos);
 if(!http_data_is_directory_listing(finfos.data))
  {
  arg_add_value(directory_listing, finfos.name, ARG_INT, sizeof(int), (void *)1);
  directory_listings++;
  report_length_listings+=strlen(finfos.name)+5;
  }
 }

 else if(finfos.error == 404)
 {
  char * parent = arg_get_value(infos, "parent");
  char * copy = estrdup(finfos.name);
  char * t = strrchr(copy, '/');
  int i;
  
  if(copy)
  {
  for(i=0;i<strlen(copy);i++)copy[i]=tolower(copy[i]);
  if(t && !strncmp(t+1, "index", 5))
  	{
         t[1]=0;
         if(!arg_get_value(links, copy))arg_add_value(links, copy, ARG_ARGLIST, -1, infos);
        }
  else free(copy);
  
  arg_add_value(error_404, finfos.name, ARG_STRING, strlen(parent), parent);
  report_length_404 += strlen(finfos.name) + strlen(parent) + 20;
  errors_404++;
  
 }
 arg_set_value(infos, "error_code", sizeof(int),(void *)finfos.error);
 free(finfos.data);
 }
 }
 l = l->next;
 }
 
 
 if(errors_404)
 {
  char * report = emalloc(report_length_404 + 255);
  struct arglist* e = error_404;
  sprintf(report, "\
The server could not find the following referenced\n\
local HTML links : \n");
  while(e && e->next)
  {
   sprintf(report, "%s%s (from %s)\n", report, e->name, (char *)e->value);
   e = e->next;
  }
  
  post_info(env, 80, report);
  free(report);
 }
 
 if(directory_listings)
 {
  char * report = emalloc(report_length_listings + 255);
  struct arglist * e = directory_listing;
  sprintf(report, "\
The server returned a file listings for the following\n\
directories which have no index : \n");
  while(e&& e->next)
  {
   sprintf(report, "%s%s\n", report, e->name);
   e = e->next;
  }
  post_info(env, 80, report);
  free(report);
 }
 
 l = links;
 while( l && l->next)
 {
  arg_add_value(export, l->name, ARG_INT, sizeof(int),(void *)1);
  l = l->next;
 }
 /* plug_set_key(env, "www/links", ARG_ARGLIST, export); */
 arg_free_all(directory_listing);
 arg_free_all(error_404);
/* arg_free_all(links); */
 return(0);
}

int http_get_file_info(env, file, server, html, length, error)
 struct arglist * env;
 char * file;
 char * server;
 int * html;
 long * length;
 int * error;
{
  int soc = open_sock_tcp(env, 80);
  char * buf = emalloc(1024+strlen(file)+strlen(server));
  char * t = emalloc(255);
  sprintf(buf, "HEAD %s HTTP/1.0\nUser-Agent: Nessus\nHost: %s\n\n", file, server);
  send(soc, buf, strlen(buf), 0);
  bzero(buf, 1024+strlen(file)+strlen(server));
  recv_line(soc, buf, 1024);
  sscanf(buf, "%s %d %s", t,error, t);
  free(t);
  if((*error == 200)||(*error == 301))
  {
   int i;
   int got_size = 0;
   int got_type = 0;
   while(!got_size || !got_type)
   {
   recv_line(soc, buf, 1024);
   if(strlen(buf)<2)break;
   for(i=0;i<strlen(buf);i++)buf[i]=tolower(buf[i]);
   if(!strncmp(buf, "content-type", 12))
    {
    if(strstr(buf, "text/html"))*html = 1;
    else *html = 0;
    got_type = 1;
    }
   if(!strncmp(buf, "content-length", 14))
    {
    t = strchr(buf, ' ');
    if(t)*length = atol(t+1);
    else *length = 0;
    got_size = 1;
    }
   }
  if(!got_size)*length = 2000;
  }
 shutdown(soc, 2);
 close(soc);
 free(buf);
 return(0);
}


int 
http_get_path_from_file(file, path)
 char * file;
 char ** path;
{
 char * ret;
 char * t;
 
 if(!file)return(1);
 ret = estrdup(file);
 t = strrchr(ret, '/');
 if(t)t[1]=0;
 else {
  if(ret)free(ret);
  ret = emalloc(2);
  ret[0]='/';
  }
 *path = ret;
 return(0);
}

int http_link_get_hostname(link, hostname)
 char * link;
 char ** hostname;
{
 char * l;
 char * start;
 char * end;
 char * ret;
 int i;
 
 if(!link)return(1);
 l = estrdup(link);
 for(i=0;i<strlen(l);i++)l[i]=tolower(l[i]);
 if(strncmp(l, "http://", 7))return(1);
 start = l+7;
 end = strchr(l, '/');
 if(end)end[0]=0;
 ret = estrdup(start);
 free(l);
 *hostname = ret;
 return(0);
}





int http_get_file(env, file_infos)
 struct arglist * env;
 struct file_infos * file_infos;
{
 int size = 0;
 int error = 0;
 int html = 0;
 char * buf;
 
 http_get_file_info(env, file_infos->name, file_infos->hostname, &html, &size, &error);
 if(error==301)
 {
  if(file_infos->name[strlen(file_infos->name)-1]!='/')
  {
   char * new = emalloc(strlen(file_infos->name)+2);
   sprintf(new, "%s/", file_infos->name);
   http_get_file_info(env, new,file_infos->hostname, &html, &size, &error );
   if(error==200){
    free(file_infos->name);
    file_infos->name = new;
   }
  }
 }
 file_infos->size = size;
 file_infos->html = html;
 file_infos->error = error;
 if(html && (error==200))
 {
  int soc;
  ssize_t b = 0;
  int ok = 0;
  soc = open_sock_tcp(env, 80);
  buf = emalloc(size+255);
  sprintf(buf, "GET %s HTTP/1.0\nHost: %s\nConnection: Keep-Alive\n\n", file_infos->name,file_infos->hostname);
  send(soc, buf, strlen(buf), 0);
  bzero(buf, size+255);
  if(size != 2000)while(b < size)b+= recv(soc, &buf[b], size+255, 0);
  else b = recv(soc, buf, size+255,0);
  file_infos->data = buf;
  shutdown(soc, 2);
  close(soc);
 }
return(0);
}

int http_extract_links(links, path, buffer, finfos)
 struct arglist * links;
 char * path;
 char * buffer;
 struct file_infos * finfos;
{
 int i=0;
 int j=0;
 char * substr;
 char * substr_2;
 char * ptr;
 
 while(buffer[i]!='<'){i++;if(i>=strlen(buffer))return(-1);}
 j = i;
 while(buffer[j]!='>'){j++;if(j>strlen(buffer))return(-1);}
 buffer[j]=0;
 substr = estrdup(&(buffer[i]));
 substr_2 = estrdup(substr);
 buffer[j]='>';
 
 for(i=0;i<strlen(substr);i++)substr[i]=tolower(substr[i]);
 if(!(ptr=strstr(substr, "href")))
 {
  free(substr);free(substr_2);return(http_extract_links(links, path, &buffer[j], finfos));
 }
 else
 {
  char * link_lower;
  char * link_orig;
  char * link = NULL;
  char * v, *t,*u;
  char * nu_path;
  int r;
  v = strchr(ptr, '"');
  if(!v)
  {
   free(substr);free(substr_2);return(http_extract_links(links, path, &buffer[j], finfos));
  }
  t = strchr(v+1, '"');
  if(!t){
   free(substr);free(substr_2);return(http_extract_links(links, path, &buffer[j], finfos));
   }
  t[0]=0;
  link_lower = estrdup(v+1);
  t[0]='"';
  u = (substr_2 + (strlen(substr_2) - strlen(v)));
  t = strchr(u+1, '"');
  if(t)t[0]=0;
  link_orig = estrdup(u+1);
  free(substr);
  free(substr_2);
  
  nu_path = estrdup(path);
  http_remove_double_dot(&link_orig, nu_path);
  if((r=http_filter_link(link_lower))>=0)
  {
   if(!r)
   {
   if(link_orig[0]!='/')
   {
    link = emalloc(strlen(nu_path)+strlen(link_orig)+1);
    sprintf(link, "%s%s", nu_path, link_orig);
   }
   else link = estrdup(link_orig);
   }
   else
   {
    char * u_hostname = NULL;
    
    http_link_get_hostname(link_orig, &u_hostname);
    if(u_hostname)
    {
     if(!strcasecmp(u_hostname, finfos->hostname))link = estrdup(link_orig+7+strlen(u_hostname));
     if(link && !strlen(link))link = estrdup("/");
    }
   }
   free(link_orig);
   free(link_lower);
   free(nu_path);
   if(link)
   {
   if(arg_get_type(links, link)<0)
   {
    struct arglist * infos = emalloc(sizeof(struct arglist));
    char * parent = estrdup(finfos->name);
    arg_add_value(infos, "parent", ARG_STRING, strlen(parent), parent);
    arg_add_value(infos, "error_code", ARG_INT, sizeof(int), 0);
    arg_add_value(links, link, ARG_ARGLIST, -1, infos);
   }
 }
 }
 }
return(http_extract_links(links,path, &buffer[j], finfos));
}
  
 
int http_filter_link(link)
 char * link;
{
 if(link[0]=='#')return(-1);
 if(!strncmp(link, "http://",7))return(1);
 if(strchr(link, ':'))return(-1);
 return(0);
}
 

void http_remove_double_dot(link, path)
 char ** link;
 char * path;
{
 char * l = *link;
 char * p = path;
 char * t;
 char * old;
 if(!strncmp(l, "../", 3))
  {
   t = strrchr(p, '/');
   if(t)t[0]=0;
   t = strrchr(p, '/');
   if(t)t[1]=0;
   l+=3;
   old = *link;
   *link = estrdup(l);
   free(old);
   http_remove_double_dot(link, path);
  }
} 

int http_data_is_directory_listing(data)
 char * data;
{
 char * buf = emalloc(strlen(data)+1);
 long i;
 int ret;
 for(i=0;i<strlen(data);i++)buf[i]=tolower(data[i]);
 if(strstr(buf, "<title>index of /"))ret=0;
 else ret=1;
 efree(&buf);
 return(ret);
}



